using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;

// HornsEditor v1.0
//
// Controls the horns shader from The Cult Assets

namespace HornsEditorPlugin
{
    public class HornsEditor : MVRScript
    {
		public JSONStorableFloat HornsDissolve;
		public JSONStorableColor HornsColor;

		private Transform ObjectRoot;
		private Renderer ObjectRenderer;
		
		private JSONStorableString _loadingInfos;
		private JSONStorableString _helpText;
				
		// Security and initialization
		private bool editorInitialized = false;
		private int initLoops = 0;
		private int maxInitLoops = 3600;
		private bool cuaIsValid = false;
		
		// Spacers
		private UIDynamic spacer0;
		private UIDynamic spacer1;
		private UIDynamic spacer2;
		
		public override void Init()
        {			 
            try
            {
				if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("Please add HornsEditor on a compatible Custom Unity Asset.");
                    return;
                }

				HSVColor hsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
				
				HornsDissolve = new JSONStorableFloat("Horns Dissolve", 0, HornsDissolveCallback, 0f, 1f);
				HornsColor = new JSONStorableColor("Horns Color", hsvc, HornsColorCallback);
				
				// Registering JSONStorables			
				RegisterFloat(HornsDissolve);
				RegisterColor(HornsColor);
				
				_helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>Horns Editor</b></size></color>\n\n" + 
					"<color=#333>" +
					"This plugin allows you to control the dissolve effect and color of the horns." +
					"</color>"
				);
				
				// Creating the default UI that will inform you that the thing is loading
				CreateLoadingUI();
				
			}
            catch(Exception e)
            {
                SuperController.LogError("HornsEditor - Exception caught: " + e);
            }
		}
		
		void Start()
        {
			
        }
		
		// Used for the initialization of the editor.
		// When it is initialized, there's nothing done every frame
		void Update()
        {
			// If the editor is not initialized, we're gonna try to do it
			// I'm putting 3600 tries ( at 60fps it is 1min )
			if( initLoops < maxInitLoops && editorInitialized == false ) {
				initLoops++;
				// If our CUA is loading and everything is ok, we can remove the loading UI and Init the final UI
				if (InitEditor() == true)
				{
					RemoveLoadingUI();
					cuaIsValid = true; // Our CUA is initialized, we consider it as valid
					
					HornsDissolveCallback(HornsDissolve.val);
					HornsColorCallback(HornsColor);
					
					StartCoroutine("CheckCUAStatus"); // We're gonna check if the CUA is valid every 5 secs
				}
			}
			
			// Triggering an error if we have reached our limit of tries
			if( initLoops == maxInitLoops && editorInitialized == false ) {
				initLoops++; // To avoid spamming the log with the error
			}
		}
		
		void OnDisable() {

		}
		
		private bool InitEditor() {
			// Updating our loading UI
			UpdateLoadingUI();

			if( editorInitialized == true ) return true;
			
			ObjectRoot = getObjectRoot( containingAtom.reParentObject, "hornsMesh" );
			if( ObjectRoot == null )
			{
				return false;
			}
			
			ObjectRenderer = ObjectRoot.GetComponent<Renderer>();
			if( ObjectRenderer == null )
			{
				return false;
			}
			
			editorInitialized = true;
			CreateCUAUI();
			
			return true;

		}
		
		// The function creating the loading UI before everything is ready
		private void CreateLoadingUI()
		{
			_loadingInfos = new JSONStorableString("Loading Infos", "");
			UIDynamic stateinfosTextfield = CreateTextField(_loadingInfos, false);
			stateinfosTextfield.height = 450.0f;
		}

		// The function clearing the loading UI
		private void RemoveLoadingUI()
		{
			// Removing the loading text field on the left of the UI
			RemoveTextField(_loadingInfos);
		}

		// The function updating the loading UI while we're either waiting for the CUA to load, or waiting for the root object and everything to be found
		private void UpdateLoadingUI()
		{
			var loadingText = "<color=#000><b>Custom Unity Asset loading...</b></color>\n" + (initLoops/60) + "/" + (maxInitLoops/60) + "secs";
			if (initLoops == maxInitLoops)
			{
				loadingText += "\n\n<color=#333><b>UNABLE TO LOAD YOUR CUA :</b>\n";
				loadingText += "- Your CUA may not be compatible with this script\n";
				loadingText += "- Your CUA may take too long to load</color>";
				loadingText += "\n\nSelect another CUA and reload the script, or just reload the script to try again.";
			}
			_loadingInfos.val = loadingText;
		}
		
		// The function creating the UI when the CUA has been invalidated
		private void CreateCUAInvalidatedUI()
		{
			JSONStorableString _invalidatedCUA = new JSONStorableString("Invalidated CUA", "");
			UIDynamic stateinfosTextfield = CreateTextField(_invalidatedCUA, false);
			stateinfosTextfield.height = 150.0f;
			_invalidatedCUA.val = "<color=#000><b>Custom Unity Asset invalidated.</b></color>\n" +
			                      "You have changed the CUA in the <b>Asset</b> tab, the script is no longer valid. Please reload it or change it.";
		}

		// Function creating the whole CUA UI
		private void CreateCUAUI()
		{
			CreateSlider(HornsDissolve);
			CreateColorPicker(HornsColor);
			
			spacer0 = CreateSpacer();
			spacer1 = CreateSpacer();		
			spacer2 = CreateSpacer();
			
		
			UIDynamic helpTextfield = CreateTextField(_helpText, true);
			helpTextfield.height = 600.0f;
		}
		
		// Function clearing the whole CUA UI
		private void ClearCUAUI()
		{
			RemoveSlider(HornsDissolve);
			RemoveColorPicker(HornsColor);
			
			RemoveSpacer(spacer0);
			RemoveSpacer(spacer1);		
			RemoveSpacer(spacer2);
			
			RemoveTextField(_helpText);
		}
		
		// Recursive function searching for a transform containing the string set in objectNameSearchString
		// It is used the first time for the initialization of the script
		// This script NEEDS to use a string "containing" your original because the cloning of the original asset will add "Clone" behind it.
		private Transform getObjectRoot( Transform parent, string rootName ) {
			foreach( Transform child in parent )
			{
				if( child.name == rootName ) {
					return child;
				} else {
					Transform childSearch = getObjectRoot(child,rootName);
					if (childSearch != null) return childSearch;
				}
			}
			
			return null;
		}

		// **************************
		// Callbacks & similar
		// **************************
		protected void HornsDissolveCallback(float amount) {
			// The "bool enabled" is useless, I'm gonna use the value of the float directly
			if (ObjectRoot != null)
			{
				ChangeHornsDissolve(amount);
			}
		}
		
		protected void HornsColorCallback(JSONStorableColor notused) {
			if (ObjectRoot != null)
			{
				Color newColor = new Color(HornsColor.colorPicker.currentColor.r, HornsColor.colorPicker.currentColor.g, HornsColor.colorPicker.currentColor.b, 1f);
				ObjectRenderer.material.SetColor("_Color",newColor);
			}
		}
		
		// **************************
		// Functions
		// **************************
		protected void ChangeHornsDissolve(float amount) {
			if( ObjectRenderer == null ) return;
			ObjectRenderer.material.SetFloat("_DissolveAmount",amount);
		}
		
		// **************************
		// COROUTINES
		// **************************
		
		// Routine checking every 5secs if our CUA is still valid, this will invalidate the UI if the user change the CUA in the asset tab
		private IEnumerator CheckCUAStatus()
		{
			while (cuaIsValid)
			{
				// Our object root is null, it means our CUA is invalid
				if (ObjectRoot == null)
				{
					// We invalidate it so that the coroutine stops by itself, remove the UI and show the UI to alert the user
					cuaIsValid = false;
					ClearCUAUI();
					CreateCUAInvalidatedUI();
				}
				yield return new WaitForSeconds(5.0f);
			}
		}

		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Cleaning modified native VAM assets
			cuaIsValid = false;
		}

		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}


	}
}