using UnityEngine;
using UnityEngine.UI;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;

// VFX/Hero Powers Editor v1.0
//
// Allows modifications and control of hero powers

namespace VFXEditorPlugin
{
    public class VFXEditor : MVRScript
    {
		private JSONStorableStringChooser copyParamsChoice;
		private string copyParamsAtomsListDefault = "Select a CUA";
		private List<string> copyParamsAtomsList;
		private List<JSONStorable> copyParamsAtomsStorables;

		public JSONStorableColor VFXColorBase;
		public JSONStorableFloat VFXColorBaseOpacity;
		public JSONStorableColor VFXColor01;
		public JSONStorableColor VFXColor02;
		public JSONStorableColor VFXColor03;
		public JSONStorableColor VFXColor04;
		public JSONStorableColor VFXColor05;
		public JSONStorableFloat VFXColor01Pos;
		public JSONStorableFloat VFXColor02Pos;
		public JSONStorableFloat VFXColor03Pos;
		public JSONStorableFloat VFXColor04Pos;
		public JSONStorableFloat VFXColor05Pos;
			
		public JSONStorableFloat VFXSimSpeed;
		public JSONStorableFloat VFXLifetimeMin;
		public JSONStorableFloat VFXLifetimeMax;
		public JSONStorableFloat VFXTSAStartFrame;
		public JSONStorableFloat VFXBaseSizeMin;
		public JSONStorableFloat VFXBaseSizeMax;

		public JSONStorableFloat VFXSizeXMin;
		public JSONStorableFloat VFXSizeXMax;
		public JSONStorableFloat VFXSizeYMin;
		public JSONStorableFloat VFXSizeYMax;
		public JSONStorableFloat VFXSizeZMin;
		public JSONStorableFloat VFXSizeZMax;
		
		public JSONStorableFloat VFXMaxParticles;
		public JSONStorableFloat VFXRateMin;
		public JSONStorableFloat VFXRateMax;
		public JSONStorableFloat VFXEmitterSizeX;
		public JSONStorableFloat VFXEmitterSizeY;
		public JSONStorableFloat VFXEmitterSizeZ;
		
		public JSONStorableFloat VFXGravity;
		public JSONStorableFloat VFXOrbitalXMin;
		public JSONStorableFloat VFXOrbitalXMax;
		public JSONStorableFloat VFXOrbitalYMin;
		public JSONStorableFloat VFXOrbitalYMax;
		public JSONStorableFloat VFXOrbitalZMin;
		public JSONStorableFloat VFXOrbitalZMax;
		
		// Transforms
		private Transform ObjectRoot;
		private Transform VFXRoot01;
		private Transform VFXEmitterPreview;
		
		// Additional UI
		private JSONStorableBool showHelpToggle;
		private JSONStorableBool editColorsToggle;
		private JSONStorableBool showEmitterToggle;
		private JSONStorableString _loadingInfos;
		private JSONStorableString _helpText;

		// VFX Components
		private ParticleSystem prtComponent;
		private ParticleSystemRenderer prtRendererComponent;
				
		// Security and initialization
		private bool editorInitialized = false;
		private int initLoops = 0;
		private int maxInitLoops = 3600;
		private bool cuaIsValid = false;
		private bool hasEmitterPreview = false;
		private string lastErrorMsg = "";
		
		// Spacers and color UIs		
		private UIDynamic[] helpUIElements;
		private UIDynamic[] colorUIElements;
		private List<UIDynamicColorPicker> allUIColorPickers;
		private List<UIDynamicSlider> allUISliders;
		private List<UIDynamicTextField> allUITextFields;
		private List<UIDynamicButton> allUIButtons;
		
		private List<UIDynamic> allUISpacers;
		
		// Default storable values
		private float LifetimeMinDef = 0.1f;
		private float LifetimeMaxDef = 0.1f;
		
		private float BaseSizeMinDef = 1f;
		private float BaseSizeMaxDef = 1f;
		private float BaseSizeXMinDef = 0.06f;
		private float BaseSizeXMaxDef = 0.06f;
		private float BaseSizeYMinDef = 0.06f;
		private float BaseSizeYMaxDef = 0.06f;
		private float BaseSizeZMinDef = 1f;
		private float BaseSizeZMaxDef = 1f;
		
		private float MaxParticlesDef = 15000f;
		private float RateMinDef = 6500f;
		private float RateMaxDef = 8000f;
		
		private float GravityDef = -0.01f;

		
		public override void Init()
        {			 
            try
            {
				if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("Please add the ember editor on a compatible Custom Unity Asset.");
                    return;
                }

				// UI and storables
				showHelpToggle = new JSONStorableBool("Show help", false, showHelpToggleCallback){ isStorable = false };
				editColorsToggle = new JSONStorableBool("Edit colors", false, editColorsToggleCallback){ isStorable = false };
				showEmitterToggle = new JSONStorableBool("Show Emitter", false, showEmitterToggleCallback){ isStorable = false };
				
				_helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>VFX Editor Help</b></size></color>\n\n" + 
					"<color=#333>" +
					"This plugin allows you to customize all CUAs VFXs from The Cult.\n\n" +
					"It can impact performances if you're using a high particle count. Be sure to be careful for realtime scene targeting VR.\n\n" +
					"Use the controls on the left to edit the appearance and behavior of the VFX.\n\n" +
					"Use the controls below this box to either copy or transfer parameters from other CUA in the current scene.\n\n" +
					"Some VFX are \"burst\" effects, changing the min/max rate will not change anything, just adjust the max particle count.\n\n" +
					"Select your target Atom in the scene with the dropdown named <b>Atoms</b>, then :\n\n" +
					"<b>Copy</b> will grab the settings from another CUA in the scene and set them in the current CUA.\n\n" +
					"<b>Transfer</b> will send settings from the current CUA to another CUA in the scene.\n\n" +
					"</color>"
				);
				
				copyParamsAtomsList = new List<string>();
				copyParamsAtomsStorables = new List<JSONStorable>();
				copyParamsChoice = new JSONStorableStringChooser("atomslist", copyParamsAtomsList, copyParamsAtomsListDefault, "Atoms", (string choice)=>
				{}){ isStorable = false };


				HSVColor hsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
				
				VFXColorBase = new JSONStorableColor("Base Color", hsvc, VFXColorCallback);
				VFXColorBaseOpacity = new JSONStorableFloat("Opacity", 1f, 0f, 1f);
				VFXColorBaseOpacity.setCallbackFunction += (val) => { VFXColorBaseOpacity.valNoCallback = val; VFXColorCallback(VFXColor01); }; // This is meant to avoid creating a specific callback for these sliders
				VFXColor01 = new JSONStorableColor("Color 01", hsvc, VFXColorCallback);
				VFXColor02 = new JSONStorableColor("Color 02", hsvc, VFXColorCallback);
				VFXColor03 = new JSONStorableColor("Color 03", hsvc, VFXColorCallback);
				VFXColor04 = new JSONStorableColor("Color 04", hsvc, VFXColorCallback);
				VFXColor05 = new JSONStorableColor("Color 05", hsvc, VFXColorCallback);
				
				VFXColor01Pos = new JSONStorableFloat("Color 01 position", 0f, 0f, 1f);
				VFXColor01Pos.setCallbackFunction += (val) => { VFXColor01Pos.valNoCallback = val; VFXColorCallback(VFXColor01); };
				VFXColor02Pos = new JSONStorableFloat("Color 02 position", 0.25f, 0f, 1f);
				VFXColor02Pos.setCallbackFunction += (val) => { VFXColor02Pos.valNoCallback = val; VFXColorCallback(VFXColor01); };
				VFXColor03Pos = new JSONStorableFloat("Color 03 position", 0.5f, 0f, 1f);
				VFXColor03Pos.setCallbackFunction += (val) => { VFXColor03Pos.valNoCallback = val; VFXColorCallback(VFXColor01); };
				VFXColor04Pos = new JSONStorableFloat("Color 04 position", 0.75f, 0f, 1f);
				VFXColor04Pos.setCallbackFunction += (val) => { VFXColor04Pos.valNoCallback = val; VFXColorCallback(VFXColor01); };
				VFXColor05Pos = new JSONStorableFloat("Color 05 position", 1f, 0f, 1f);
				VFXColor05Pos.setCallbackFunction += (val) => { VFXColor05Pos.valNoCallback = val; VFXColorCallback(VFXColor01); };
							
				VFXSimSpeed = new JSONStorableFloat("Simulation Speed", 1f, VFXBaseCallback, 0.01f, 3f);
				VFXLifetimeMin = new JSONStorableFloat("Particle Lifetime Min", LifetimeMinDef, VFXBaseCallback, 0.01f, 10f);
				VFXLifetimeMax = new JSONStorableFloat("Particle Lifetime Max", LifetimeMaxDef, VFXBaseCallback, 0.01f, 10f);
				VFXTSAStartFrame = new JSONStorableFloat("Particle Type", 14f, 0f, 19f);
				VFXTSAStartFrame.setCallbackFunction += (val) => { VFXTSAStartFrame.valNoCallback = Mathf.Round(val); VFXBaseCallback(VFXSimSpeed); };

				VFXBaseSizeMin = new JSONStorableFloat("Particle Size Min", BaseSizeMinDef, VFXSizeCallback, 0.01f, 10f);
				VFXBaseSizeMax = new JSONStorableFloat("Particle Size Max", BaseSizeMaxDef, VFXSizeCallback, 0.01f, 10f);				
				VFXSizeXMin = new JSONStorableFloat("Particle Size X Min", BaseSizeXMinDef, VFXSizeCallback, 0.01f, 10f);
				VFXSizeXMax = new JSONStorableFloat("Particle Size X Max", BaseSizeXMaxDef, VFXSizeCallback, 0.01f, 10f);
				VFXSizeYMin = new JSONStorableFloat("Particle Size Y Min", BaseSizeYMinDef, VFXSizeCallback, 0.01f, 10f);
				VFXSizeYMax = new JSONStorableFloat("Particle Size Y Max", BaseSizeYMaxDef, VFXSizeCallback, 0.01f, 10f);
				VFXSizeZMin = new JSONStorableFloat("Particle Size Z Min", BaseSizeZMinDef, VFXSizeCallback, 1f, 1f);
				VFXSizeZMax = new JSONStorableFloat("Particle Size Z Max", BaseSizeZMaxDef, VFXSizeCallback, 1f, 1f);
				
				VFXMaxParticles = new JSONStorableFloat("Max Particles", MaxParticlesDef, VFXEmitterCallback, 1f, 100000f);
				VFXRateMin = new JSONStorableFloat("Particles Rate Min", RateMinDef, VFXEmitterCallback, 1f, 25000f);
				VFXRateMax = new JSONStorableFloat("Particles Rate Max", RateMaxDef, VFXEmitterCallback, 1f, 25000f);
				
				VFXEmitterSizeX = new JSONStorableFloat("Emitter Size X", 1f, VFXEmitterCallback, 0.1f, 5f);
				VFXEmitterSizeY = new JSONStorableFloat("Emitter Size Y", 1f, VFXEmitterCallback, 0.1f, 5f);
				VFXEmitterSizeZ = new JSONStorableFloat("Emitter Size Z", 1f, VFXEmitterCallback, 0.1f, 5f);
				
				VFXGravity = new JSONStorableFloat("Gravity", GravityDef, VFXMotionCallback, -0.05f, 0.05f, false);
				VFXOrbitalXMin = new JSONStorableFloat("Orbital X Min", 0f, VFXMotionCallback, -5f, 5f);
				VFXOrbitalXMax = new JSONStorableFloat("Orbital X Max", 0f, VFXMotionCallback, -5f, 5f);
				VFXOrbitalYMin = new JSONStorableFloat("Orbital Y Min", 0f, VFXMotionCallback, -5f, 5f);
				VFXOrbitalYMax = new JSONStorableFloat("Orbital Y Max", 0f, VFXMotionCallback, -5f, 5f);
				VFXOrbitalZMin = new JSONStorableFloat("Orbital Z Min", 0f, VFXMotionCallback, -5f, 5f);
				VFXOrbitalZMax = new JSONStorableFloat("Orbital Z Max", 0f, VFXMotionCallback, -5f, 5f);
				
				// Registering JSONStorables
				RegisterStringChooser(copyParamsChoice);
				
				RegisterColor(VFXColorBase);
				RegisterFloat(VFXColorBaseOpacity);
				RegisterColor(VFXColor01);
				RegisterColor(VFXColor02);
				RegisterColor(VFXColor03);
				RegisterColor(VFXColor04);
				RegisterColor(VFXColor05);
				RegisterFloat(VFXColor01Pos);
				RegisterFloat(VFXColor02Pos);
				RegisterFloat(VFXColor03Pos);
				RegisterFloat(VFXColor04Pos);
				RegisterFloat(VFXColor05Pos);
				
				RegisterFloat(VFXSimSpeed);
				RegisterFloat(VFXLifetimeMin);
				RegisterFloat(VFXLifetimeMax);
				RegisterFloat(VFXTSAStartFrame);
				
				RegisterFloat(VFXSizeXMin);
				RegisterFloat(VFXSizeXMax);
				RegisterFloat(VFXSizeYMin);
				RegisterFloat(VFXSizeYMax);
				RegisterFloat(VFXSizeZMin);
				RegisterFloat(VFXSizeZMax);
				
				RegisterFloat(VFXMaxParticles);
				RegisterFloat(VFXRateMin);
				RegisterFloat(VFXRateMax);
				
				RegisterFloat(VFXEmitterSizeX);
				RegisterFloat(VFXEmitterSizeY);
				RegisterFloat(VFXEmitterSizeZ);
				
				RegisterFloat(VFXGravity);
				RegisterFloat(VFXOrbitalXMin);
				RegisterFloat(VFXOrbitalXMax);
				RegisterFloat(VFXOrbitalYMin);
				RegisterFloat(VFXOrbitalYMax);
				RegisterFloat(VFXOrbitalZMin);
				RegisterFloat(VFXOrbitalZMax);
				
				RegisterBool(showHelpToggle);
				// Creating the default UI that will inform you that the thing is loading
				CreateLoadingUI();
				
			}
            catch(Exception e)
            {
                SuperController.LogError("Ember 01 editor - Exception caught: " + e);
            }
		}
		
		void Start()
        {
			
        }
		
		// Used for the initialization of the editor.
		// When it is initialized, there's nothing done every frame
		void Update()
        {
			// If the editor is not initialized, we're gonna try to do it
			// I'm putting 3600 tries ( at 60fps it is 1min )
			if( initLoops < maxInitLoops && editorInitialized == false ) {
				initLoops++;
				// If our CUA is loading and everything is ok, we can remove the loading UI and Init the final UI
				if (InitEditor() == true)
				{
					RemoveLoadingUI();
					cuaIsValid = true; // Our CUA is initialized, we consider it as valid
					
					showEmitterToggleCallback(false);
					VFXColorCallback(VFXColor01);
					VFXBaseCallback(VFXLifetimeMin);
					VFXSizeCallback(VFXSizeXMin);
					VFXEmitterCallback(VFXEmitterSizeX);
					VFXMotionCallback(VFXOrbitalXMin);
					
					StartCoroutine("CheckCUAStatus"); // We're gonna check if the CUA is valid every 5 secs
				}
			}
			
			// Triggering an error if we have reached our limit of tries
			if( initLoops == maxInitLoops && editorInitialized == false ) {
				initLoops++; // To avoid spamming the log with the error
			}
		}
		
		void OnDisable() {

		}
		
		private bool InitEditor() {
			// Updating our loading UI
			UpdateLoadingUI();

			if( editorInitialized == true ) return true;
			
			// Getting main components
			ObjectRoot = getVFXRoot( containingAtom.reParentObject, "root_vfx" );
			if( ObjectRoot == null ) {
				lastErrorMsg = "This CUA does not have a compatible root object";
				return false;
			}
			
			VFXRoot01 = getVFXRoot( ObjectRoot, "vfx_01" );
			if( VFXRoot01 == null ) {
				lastErrorMsg = "There is no vfx_01 object in this VFX CUA";
				return false;
			}
			
			VFXEmitterPreview = getVFXRoot( ObjectRoot, "_emitter_preview" );
			if( VFXEmitterPreview != null ) hasEmitterPreview = true;
			
			prtComponent = VFXRoot01.GetComponent<ParticleSystem>();
			if( prtComponent == null ) {
				lastErrorMsg = "Unable to get the particle system on this VFX CUA";
				return false;
			}
			
			prtRendererComponent = VFXRoot01.GetComponent<ParticleSystemRenderer>();
			if( prtRendererComponent == null ) {
				lastErrorMsg = "Unable to get the particle system renderer on this VFX CUA";
				return false;
			}

			InitCopyParamsList();
			
			editorInitialized = true;
				
			CreateCUAUI();
			
			return true;
		}
		
		private void InitCopyParamsList() {
			foreach(Atom at in SuperController.singleton.GetAtoms() ) {
				if( at.type == "CustomUnityAsset" ) {
					JSONStorable vfxEditor = at.GetStorableIDs().Select(at.GetStorableByID).FirstOrDefault(s => s.IsColorJSONParam("Color 01"));
					if( vfxEditor != null ) {
						if( at.name != containingAtom.name ) {						
							copyParamsAtomsList.Add( at.name );
							copyParamsAtomsStorables.Add( vfxEditor );
						}
					}
				}
			}
		}
		
		// The function creating the loading UI before everything is ready
		private void CreateLoadingUI()
		{
			_loadingInfos = new JSONStorableString("Loading Infos", "");
			UIDynamic stateinfosTextfield = CreateTextField(_loadingInfos, false);
			stateinfosTextfield.height = 450.0f;
		}

		// The function clearing the loading UI
		private void RemoveLoadingUI()
		{
			// Removing the loading text field on the left of the UI
			RemoveTextField(_loadingInfos);
		}

		// The function updating the loading UI while we're either waiting for the CUA to load, or waiting for the root object and everything to be found
		private void UpdateLoadingUI()
		{
			var loadingText = "<color=#000><b>Custom Unity Asset loading...</b></color>\n" + (initLoops/60) + "/" + (maxInitLoops/60) + "secs";
			if (initLoops == maxInitLoops)
			{
				loadingText += "\n\n<color=#333><b>UNABLE TO LOAD YOUR CUA :</b>\n";
				loadingText += "- Your CUA may not be compatible with this script\n";
				loadingText += "- Your CUA may take too long to load</color>";
				loadingText += "\n\nSelect another CUA and reload the script, or just reload the script to try again.";
				loadingText += "\n\n<b>Last error message :</b> " + lastErrorMsg;
			}
			_loadingInfos.val = loadingText;
		}
		
		// The function creating the UI when the CUA has been invalidated
		private void CreateCUAInvalidatedUI()
		{
			JSONStorableString _invalidatedCUA = new JSONStorableString("Invalidated CUA", "");
			UIDynamic stateinfosTextfield = CreateTextField(_invalidatedCUA, false);
			stateinfosTextfield.height = 150.0f;
			_invalidatedCUA.val = "<color=#000><b>Custom Unity Asset invalidated.</b></color>\n" +
			                      "You have changed the CUA in the <b>Asset</b> tab, the script is no longer valid. Please reload it or change it.";
		}

		// Function creating the whole CUA UI
		private void CreateCUAUI()
		{
			allUIColorPickers = new List<UIDynamicColorPicker>();
			allUISliders = new List<UIDynamicSlider>();
			allUITextFields = new List<UIDynamicTextField>();
			allUIButtons = new List<UIDynamicButton>();
			allUISpacers = new List<UIDynamic>();
			
			UIDynamicSlider tmpSlider;
			UIDynamicColorPicker tmpColor;
			UIDynamic tmpUID;
			
			// Create the UI depending on the VFX
			VFX_CreateUI();
			
			/** *************** COPY AND TRANSFER BUTTONS ************************* **/
			UIDynamicPopup editTriggerPopup = CreateScrollablePopup(copyParamsChoice,true);
			
			// Copy
			var tmpBtn = CreateButton("Copy Colors", true);
			tmpBtn.button.onClick.AddListener( () => { CopyColorsFromAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);
			
			tmpBtn = CreateButton("Copy base options", true);
			tmpBtn.button.onClick.AddListener( () => { CopyBaseOptionsFromAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);
			
			tmpBtn = CreateButton("Copy size options", true);
			tmpBtn.button.onClick.AddListener( () => { CopySizeOptionsFromAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);
			
			tmpBtn = CreateButton("Copy emitter options", true);
			tmpBtn.button.onClick.AddListener( () => { CopyEmitterOptionsFromAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);
			
			tmpBtn = CreateButton("Copy motion options", true);
			tmpBtn.button.onClick.AddListener( () => { CopyMotionOptionsFromAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);
			
			allUISpacers.Add(CreateSpacer(true));
			
			// Transfers
			tmpBtn = CreateButton("Transfer Colors", true);
			tmpBtn.button.onClick.AddListener( () => { TransferColorsToAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);
			
			tmpBtn = CreateButton("Transfer base options", true);
			tmpBtn.button.onClick.AddListener( () => { TransferBaseOptionsToAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);
			
			tmpBtn = CreateButton("Transfer size options", true);
			tmpBtn.button.onClick.AddListener( () => { TransferSizeOptionsToAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);
			
			tmpBtn = CreateButton("Transfer emitter options", true);
			tmpBtn.button.onClick.AddListener( () => { TransferEmitterOptionsToAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);

			tmpBtn = CreateButton("Transfer motion options", true);
			tmpBtn.button.onClick.AddListener( () => { TransferMotionOptionsToAtom( copyParamsChoice.val ); });
			allUIButtons.Add(tmpBtn);
			
			// Toggling UI elements off by default
			showHelpToggleCallback(false);
			editColorsToggleCallback(false);
		}
		
		// Function clearing the whole CUA UI
		private void ClearCUAUI()
		{
			foreach( UIDynamicColorPicker uiEl in allUIColorPickers ) {
				RemoveColorPicker(uiEl);
			}
			
			foreach( UIDynamicSlider uiEl in allUISliders ) {
				RemoveSlider(uiEl);
			}
			
			foreach( UIDynamicTextField uiEl in allUITextFields ) {
				RemoveTextField(uiEl);
			}
			
			foreach( UIDynamic uiEl in allUISpacers ) {
				RemoveSpacer(uiEl);
			}
			
			foreach( UIDynamicButton uiEl in allUIButtons ) {
				RemoveButton(uiEl);
			}

			RemoveTextField(_helpText);
			RemoveToggle(showHelpToggle);
			
			RemovePopup(copyParamsChoice);
		}
				
		private Transform getVFXRoot( Transform parent, string searchName ) {
			foreach( Transform child in parent )
			{
				if( child.name == searchName ) {
					return child;
				} else {
					Transform childSearch = getVFXRoot(child,searchName);
					if (childSearch != null) return childSearch;
				}
			}
			
			return null;
		}


		// **************************
		// Callbacks & similar
		// **************************
		
		protected void editColorsToggleCallback(bool enabled) {
			foreach( UIDynamic uidElem in colorUIElements ) {
				if( enabled == true ) {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = false;
				} else {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = true;
				}
			}
		}
		
		protected void showHelpToggleCallback(bool enabled) {
			foreach( UIDynamic uidElem in helpUIElements ) {
				if( enabled == true ) {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = false;
				} else {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = true;
				}
			}
		}
		
		protected void showEmitterToggleCallback(bool enabled) {
			if( hasEmitterPreview == true ) {
				VFXEmitterPreview.gameObject.SetActive( showEmitterToggle.val );
			}
		}
		
		protected void VFXColorCallback(JSONStorableColor notused) {
			if (ObjectRoot != null)
			{
				var prtMain = prtComponent.main;
				prtMain.startColor = new Color(VFXColorBase.colorPicker.currentColor.r, VFXColorBase.colorPicker.currentColor.g, VFXColorBase.colorPicker.currentColor.b, VFXColorBaseOpacity.val);
				
				var col = prtComponent.colorOverLifetime;
				UnityEngine.Gradient grd = col.color.gradientMax;
				UnityEngine.Gradient newGrd = new UnityEngine.Gradient();
				
				UnityEngine.GradientColorKey[] colorKeys = new GradientColorKey[5];
				colorKeys[0].color = new Color(VFXColor01.colorPicker.currentColor.r, VFXColor01.colorPicker.currentColor.g, VFXColor01.colorPicker.currentColor.b, 1.0f);
				colorKeys[0].time = VFXColor01Pos.val;
				colorKeys[1].color = new Color(VFXColor02.colorPicker.currentColor.r, VFXColor02.colorPicker.currentColor.g, VFXColor02.colorPicker.currentColor.b, 1.0f);
				colorKeys[1].time = VFXColor02Pos.val;
				colorKeys[2].color = new Color(VFXColor03.colorPicker.currentColor.r, VFXColor03.colorPicker.currentColor.g, VFXColor03.colorPicker.currentColor.b, 1.0f);
				colorKeys[2].time = VFXColor03Pos.val;
				colorKeys[3].color = new Color(VFXColor04.colorPicker.currentColor.r, VFXColor04.colorPicker.currentColor.g, VFXColor04.colorPicker.currentColor.b, 1.0f);
				colorKeys[3].time = VFXColor04Pos.val;
				colorKeys[4].color = new Color(VFXColor05.colorPicker.currentColor.r, VFXColor05.colorPicker.currentColor.g, VFXColor05.colorPicker.currentColor.b, 1.0f);
				colorKeys[4].time = VFXColor05Pos.val;

				newGrd.SetKeys( colorKeys, grd.alphaKeys );
				
				col.color = newGrd;				
			}
		}
		
		protected void VFXBaseCallback(JSONStorableFloat notused) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.simulationSpeed = VFXSimSpeed.val;
				prtMain.startLifetime = new ParticleSystem.MinMaxCurve(VFXLifetimeMin.val, VFXLifetimeMax.val);
				
				// You're wondering why 0.03125f ?
				// for a frikkin reason behind the TSA startFrame, unity wants the actual float value of the total number of frame of the flipbook remapped to the value of 1
				// Which is not the case at all in the UI in the editor (where you input the actual frame value)
				// In this case, the flipbook is 8x4 = 32 frames, 1 / 32 = 0.03125
				var prtTsa = prtComponent.textureSheetAnimation;
				prtTsa.startFrame = 0.03125f * VFXTSAStartFrame.val;
			}
		}
				
		protected void VFXSizeCallback(JSONStorableFloat notused) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				
				prtMain.startSize = new ParticleSystem.MinMaxCurve(VFXBaseSizeMin.val, VFXBaseSizeMax.val);

				prtMain.startSizeX = new ParticleSystem.MinMaxCurve(VFXSizeXMin.val, VFXSizeXMax.val);
				prtMain.startSizeY = new ParticleSystem.MinMaxCurve(VFXSizeYMin.val, VFXSizeYMax.val);
				prtMain.startSizeZ = new ParticleSystem.MinMaxCurve(VFXSizeZMin.val, VFXSizeZMax.val);
			}
		}
		
		protected void VFXEmitterCallback(JSONStorableFloat notused) {
			if (ObjectRoot != null)
			{
				var prtShape = prtComponent.shape;
				prtShape.scale = new Vector3( VFXEmitterSizeX.val, VFXEmitterSizeY.val, VFXEmitterSizeZ.val );
				UpdateEmitterPreview();
				
				var prtMain = prtComponent.main;
				prtMain.maxParticles = (int)VFXMaxParticles.val;
				
				var prtEmission = prtComponent.emission;
				prtEmission.rate = new ParticleSystem.MinMaxCurve(VFXRateMin.val, VFXRateMax.val);
			}
		}
		
		protected void VFXMotionCallback(JSONStorableFloat notused) {
			if (ObjectRoot != null)
			{
				var prtMain = prtComponent.main;
				prtMain.gravityModifier = VFXGravity.val;
				
				var prtVol = prtComponent.velocityOverLifetime;
				prtVol.orbitalX = new ParticleSystem.MinMaxCurve(VFXOrbitalXMin.val, VFXOrbitalXMax.val);
				prtVol.orbitalY = new ParticleSystem.MinMaxCurve(VFXOrbitalYMin.val, VFXOrbitalYMax.val);
				prtVol.orbitalZ = new ParticleSystem.MinMaxCurve(VFXOrbitalZMin.val, VFXOrbitalZMax.val);
				
			}
		}
		
		// **************************
		// Functions
		// **************************
		private void VFX_CreateUI() {
			UIDynamicSlider tmpSlider;
			UIDynamicColorPicker tmpColor;
			UIDynamic tmpUID;
			
			/** *************** COLORS ************************* **/
			CreateToggle(editColorsToggle, false);
			
			colorUIElements = new UIDynamic[13];
			tmpColor = CreateColorPicker(VFXColorBase);
			allUIColorPickers.Add(tmpColor);
			colorUIElements[0] = (UIDynamic)tmpColor;
			
			tmpSlider = CreateSlider(VFXColorBaseOpacity);
			allUISliders.Add(tmpSlider);
			colorUIElements[1] = (UIDynamic)tmpSlider;
			
			tmpColor = CreateColorPicker(VFXColor01);
			allUIColorPickers.Add(tmpColor);
			colorUIElements[2] = (UIDynamic)tmpColor;
			
			tmpColor = CreateColorPicker(VFXColor02);
			allUIColorPickers.Add(tmpColor);
			colorUIElements[3] = (UIDynamic)tmpColor;
			
			tmpColor = CreateColorPicker(VFXColor03);
			allUIColorPickers.Add(tmpColor);
			colorUIElements[4] = (UIDynamic)tmpColor;
			
			tmpColor = CreateColorPicker(VFXColor04);
			allUIColorPickers.Add(tmpColor);
			colorUIElements[5] = (UIDynamic)tmpColor;
			
			tmpColor = CreateColorPicker(VFXColor05);
			allUIColorPickers.Add(tmpColor);
			colorUIElements[6] = (UIDynamic)tmpColor;
			
			
			tmpSlider = CreateSlider(VFXColor01Pos);
			allUISliders.Add(tmpSlider);
			colorUIElements[7] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(VFXColor02Pos);
			allUISliders.Add(tmpSlider);
			colorUIElements[8] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(VFXColor03Pos);
			allUISliders.Add(tmpSlider);
			colorUIElements[9] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(VFXColor04Pos);
			allUISliders.Add(tmpSlider);
			colorUIElements[10] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(VFXColor05Pos);
			allUISliders.Add(tmpSlider);
			colorUIElements[11] = (UIDynamic)tmpSlider;
			
			tmpUID = CreateSpacer();
			allUISpacers.Add(tmpUID);
			colorUIElements[12] = tmpUID;
			
			/** *************** BASE OPTIONS ************************* **/
			allUITextFields.Add(createStaticDescriptionText("Base Options","<color=#000><size=35><b>Base options</b></size></color>",false,45));

			allUISliders.Add(CreateSlider(VFXSimSpeed));
			allUISliders.Add(CreateSlider(VFXLifetimeMin));
			allUISliders.Add(CreateSlider(VFXLifetimeMax));
			allUISliders.Add(CreateSlider(VFXTSAStartFrame));

			allUISpacers.Add(CreateSpacer());

			/** *************** SIZE OPTIONS ************************* **/
			allUITextFields.Add(createStaticDescriptionText("Size Options","<color=#000><size=35><b>Size options</b></size></color>",false,45));
			
			allUISliders.Add(CreateSlider(VFXSizeXMin));
			allUISliders.Add(CreateSlider(VFXSizeXMax));
			allUISliders.Add(CreateSlider(VFXSizeYMin));
			allUISliders.Add(CreateSlider(VFXSizeYMax));
			// Slider Z is not needed, we're on a flat particle

			allUISpacers.Add(CreateSpacer());
			
			/** *************** EMITTER OPTIONS ************************* **/
			allUITextFields.Add(createStaticDescriptionText("Emitter Options","<color=#000><size=35><b>Emitter Options</b></size></color>",false,45));
			
			allUISliders.Add(CreateSlider(VFXMaxParticles));
			allUISliders.Add(CreateSlider(VFXRateMin));
			allUISliders.Add(CreateSlider(VFXRateMax));
			allUISliders.Add(CreateSlider(VFXEmitterSizeX));
			allUISliders.Add(CreateSlider(VFXEmitterSizeY));
			allUISliders.Add(CreateSlider(VFXEmitterSizeZ));
			
			allUISpacers.Add(CreateSpacer());
			
			/** *************** MOTION OPTIONS ************************* **/
			allUITextFields.Add(createStaticDescriptionText("Motion Options","<color=#000><size=35><b>Motion Options</b></size></color>",false,45));
			
			allUISliders.Add(CreateSlider(VFXGravity));
			allUISliders.Add(CreateSlider(VFXOrbitalXMin));
			allUISliders.Add(CreateSlider(VFXOrbitalXMax));
			allUISliders.Add(CreateSlider(VFXOrbitalYMin));
			allUISliders.Add(CreateSlider(VFXOrbitalYMax));
			allUISliders.Add(CreateSlider(VFXOrbitalZMin));
			allUISliders.Add(CreateSlider(VFXOrbitalZMax));
			
			CreateToggle(showHelpToggle, true);
			
			helpUIElements = new UIDynamic[1];
			helpUIElements[0] = CreateTextField(_helpText, true);
			helpUIElements[0].height = 600.0f;
			
		}
		
		protected void UpdateEmitterPreview() {
			if( hasEmitterPreview == true ) {
				VFXEmitterPreview.localScale = new Vector3( VFXEmitterSizeX.val, VFXEmitterSizeY.val, VFXEmitterSizeZ.val );
			}
		}
		
		protected void CopyColorsFromAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				VFXColorBase.val = selectedStorable.GetColorJSONParam("Base Color").val;
				VFXColorBaseOpacity.val = selectedStorable.GetFloatJSONParam("Opacity").val;
				VFXColor01.val = selectedStorable.GetColorJSONParam("Color 01").val;
				VFXColor02.val = selectedStorable.GetColorJSONParam("Color 02").val;
				VFXColor03.val = selectedStorable.GetColorJSONParam("Color 03").val;
				VFXColor04.val = selectedStorable.GetColorJSONParam("Color 04").val;
				VFXColor05.val = selectedStorable.GetColorJSONParam("Color 05").val;
				VFXColor01Pos.val = selectedStorable.GetFloatJSONParam("Color 01 position").val;
				VFXColor02Pos.val = selectedStorable.GetFloatJSONParam("Color 02 position").val;
				VFXColor03Pos.val = selectedStorable.GetFloatJSONParam("Color 03 position").val;
				VFXColor04Pos.val = selectedStorable.GetFloatJSONParam("Color 04 position").val;
				VFXColor05Pos.val = selectedStorable.GetFloatJSONParam("Color 05 position").val;
			}
		}
		
		protected void CopyBaseOptionsFromAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				VFXSimSpeed.val = selectedStorable.GetFloatJSONParam("Simulation Speed").val;
				VFXLifetimeMin.val = selectedStorable.GetFloatJSONParam("Particle Lifetime Min").val;
				VFXLifetimeMax.val = selectedStorable.GetFloatJSONParam("Particle Lifetime Max").val;
				VFXTSAStartFrame.val = selectedStorable.GetFloatJSONParam("Particle Type").val;
			}
		}
		
		protected void CopySizeOptionsFromAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				VFXSizeXMin.val = selectedStorable.GetFloatJSONParam("Particle Size X Min").val;
				VFXSizeXMax.val = selectedStorable.GetFloatJSONParam("Particle Size X Max").val;
				VFXSizeYMin.val = selectedStorable.GetFloatJSONParam("Particle Size Y Min").val;
				VFXSizeYMax.val = selectedStorable.GetFloatJSONParam("Particle Size Y Max").val;
				VFXSizeZMin.val = selectedStorable.GetFloatJSONParam("Particle Size Z Min").val;
				VFXSizeZMax.val = selectedStorable.GetFloatJSONParam("Particle Size Z Max").val;
			}
		}
		
		protected void CopyEmitterOptionsFromAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				VFXMaxParticles.val = selectedStorable.GetFloatJSONParam("Max Particles").val;
				VFXRateMin.val = selectedStorable.GetFloatJSONParam("Particles Rate Min").val;
				VFXRateMax.val = selectedStorable.GetFloatJSONParam("Particles Rate Max").val;
				VFXEmitterSizeX.val = selectedStorable.GetFloatJSONParam("Emitter Size X").val;
				VFXEmitterSizeY.val = selectedStorable.GetFloatJSONParam("Emitter Size Y").val;
				VFXEmitterSizeZ.val = selectedStorable.GetFloatJSONParam("Emitter Size Z").val;
			}
		}
		
		protected void CopyMotionOptionsFromAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				VFXGravity.val = selectedStorable.GetFloatJSONParam("Gravity").val;
				VFXOrbitalXMin.val = selectedStorable.GetFloatJSONParam("Orbital X Min").val;
				VFXOrbitalXMax.val = selectedStorable.GetFloatJSONParam("Orbital X Max").val;
				VFXOrbitalYMin.val = selectedStorable.GetFloatJSONParam("Orbital Y Min").val;
				VFXOrbitalYMax.val = selectedStorable.GetFloatJSONParam("Orbital Y Max").val;
				VFXOrbitalZMin.val = selectedStorable.GetFloatJSONParam("Orbital Z Min").val;
				VFXOrbitalZMax.val = selectedStorable.GetFloatJSONParam("Orbital Z Max").val;
			}
		}
		
		protected void TransferColorsToAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				selectedStorable.SetColorParamValue("Base Color", VFXColorBase.val);
				selectedStorable.SetFloatParamValue("Opacity", VFXColorBaseOpacity.val);
				selectedStorable.SetColorParamValue("Color 01", VFXColor01.val);
				selectedStorable.SetColorParamValue("Color 02", VFXColor02.val);
				selectedStorable.SetColorParamValue("Color 03", VFXColor03.val);
				selectedStorable.SetColorParamValue("Color 04", VFXColor04.val);
				selectedStorable.SetColorParamValue("Color 05", VFXColor05.val);
				selectedStorable.SetFloatParamValue("Color 01 position", VFXColor01Pos.val);
				selectedStorable.SetFloatParamValue("Color 02 position", VFXColor02Pos.val);
				selectedStorable.SetFloatParamValue("Color 03 position", VFXColor03Pos.val);
				selectedStorable.SetFloatParamValue("Color 04 position", VFXColor04Pos.val);
				selectedStorable.SetFloatParamValue("Color 05 position", VFXColor05Pos.val);
			}
		}
		
		protected void TransferBaseOptionsToAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				selectedStorable.SetFloatParamValue("Simulation Speed", VFXSimSpeed.val);
				selectedStorable.SetFloatParamValue("Particle Lifetime Min", VFXLifetimeMin.val);
				selectedStorable.SetFloatParamValue("Particle Lifetime Max", VFXLifetimeMax.val);
				selectedStorable.SetFloatParamValue("Particle Type", VFXTSAStartFrame.val);
			}
		}
		
		protected void TransferSizeOptionsToAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				selectedStorable.SetFloatParamValue("Particle Size X Min", VFXSizeXMin.val);
				selectedStorable.SetFloatParamValue("Particle Size X Max", VFXSizeXMax.val);
				selectedStorable.SetFloatParamValue("Particle Size Y Min", VFXSizeYMin.val);
				selectedStorable.SetFloatParamValue("Particle Size Y Max", VFXSizeYMax.val);
				selectedStorable.SetFloatParamValue("Particle Size Z Min", VFXSizeZMin.val);
				selectedStorable.SetFloatParamValue("Particle Size Z Max", VFXSizeZMax.val);
			}
		}
		
		protected void TransferEmitterOptionsToAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				selectedStorable.SetFloatParamValue("Max Particles", VFXMaxParticles.val);
				selectedStorable.SetFloatParamValue("Particles Rate Min", VFXRateMin.val);
				selectedStorable.SetFloatParamValue("Particles Rate Max", VFXRateMax.val);
				selectedStorable.SetFloatParamValue("Emitter Size X", VFXEmitterSizeX.val);
				selectedStorable.SetFloatParamValue("Emitter Size Y", VFXEmitterSizeY.val);
				selectedStorable.SetFloatParamValue("Emitter Size Z", VFXEmitterSizeZ.val);
			}
		}
		
		protected void TransferMotionOptionsToAtom( string targetAtom ) {
			JSONStorable selectedStorable = copyParamsAtomsStorables[ copyParamsAtomsList.IndexOf(targetAtom) ];
			if( selectedStorable != null ) {
				selectedStorable.SetFloatParamValue("Gravity", VFXGravity.val);
				selectedStorable.SetFloatParamValue("Orbital X Min", VFXOrbitalXMin.val);
				selectedStorable.SetFloatParamValue("Orbital X Min", VFXOrbitalXMax.val);
				selectedStorable.SetFloatParamValue("Orbital Y Min", VFXOrbitalYMin.val);
				selectedStorable.SetFloatParamValue("Orbital Y Min", VFXOrbitalYMax.val);
				selectedStorable.SetFloatParamValue("Orbital Z Min", VFXOrbitalZMin.val);
				selectedStorable.SetFloatParamValue("Orbital Z Min", VFXOrbitalZMax.val);
			}
		}


		
		// **************************
		// COROUTINES
		// **************************
		
		// Routine checking every 5secs if our CUA is still valid, this will invalidate the UI if the user change the CUA in the asset tab
		private IEnumerator CheckCUAStatus()
		{
			while (cuaIsValid)
			{
				// Our object root is null, it means our CUA is invalid
				if (ObjectRoot == null)
				{
					// We invalidate it so that the coroutine stops by itself, remove the UI and show the UI to alert the user
					cuaIsValid = false;
					ClearCUAUI();
					CreateCUAInvalidatedUI();
				}
				yield return new WaitForSeconds(5.0f);
			}
		}


		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}


		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		private UIDynamicTextField createStaticDescriptionText( string DescTitle, string DescText, bool rightSide, int fieldHeight ) {
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText);
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);

			return staticDescStringField;
		}

	}
}